import { parse as parseYaml } from "@std/yaml";
import { list } from "./main.ts";
import { normalize } from "@std/path/posix";

interface Config {
    ignore: Set<string>;
    plugins: Set<string>;

    maxCache?: number;
    debug: boolean;
    track?: boolean;
    static?: string;
    libs?: string;
    views?: string;
    notFound: string;

    defaultPage?: string;
    defaultIndex?: string;
    defaultTitle?: string;
}

function merge<T>(...objs: T[]): T {
    if (objs.length === 0) throw new Error("Expected at least one object to merge");
    let res = objs[0];

    for (const obj of objs.slice(1)) {
        if (obj == null) res = undefined as T;
        else if (Array.isArray(obj)) {
            if (Array.isArray(res)) res.push(...obj);
            else res = obj;
        }
        else if (typeof obj === "object") {
            if (typeof res !== "object" || res == null) {
                res = obj;
            }
            else {
                for (const key in obj) {
                    if (key in res) res[key] = merge(res[key], obj[key]);
                    else res[key] = obj[key];
                }
            }
        }
        else res = obj;
    }

    return res;
}

async function readConfig() {
    const objs = await Promise.all(Deno.args.map(async v => parseYaml(await Deno.readTextFile(v)) as Config));
    const res = merge(...objs);
    res.ignore = new Set(list(res.ignore).map(v => normalize(v)));
    res.plugins = new Set(list(res.plugins).map(v => normalize(v)));

    return res;
}

const config = await readConfig();

export default config;

export let isIgnored = (path: string): boolean => {
    const regexs: string[] = [];

    for (let predicate of config.ignore) {
        predicate = normalize(predicate);
        let absolute;

        if (predicate === "/") {
            isIgnored = () => true;
            return true;
        }

        if (predicate.startsWith("/")) {
            absolute = true;
            predicate = predicate.slice(1);
        }
        if (predicate.endsWith("/")) {
            predicate = predicate.slice(0, -1);
        }
        if (predicate.startsWith("./") ||
            predicate.startsWith("../") ||
            predicate == ".." ||
            predicate == ".") throw new Error(`'${predicate}' in ignore list includes relative specifiers`);

        const body = predicate.split("/")
            .filter(v => v !== "")
            .map(v => v.replace(/\*\*|[.*+?^${}()|[\]\\]/g, val => {
                if (val === "*") return "[^/]*";
                else if (val === "**") return ".*";
                else return "\\" + val;
            }))
            .join("/");

        regexs.push(absolute ? "(^/" + body + "/)" : "(/" + body + "/)");
    }

    const regex = new RegExp(regexs.join("|"));

    isIgnored = path => regex.test(normalize("/" + path + "/"));
    return isIgnored(path);
};
