import { parse as parseYaml } from "@std/yaml";
import config from "../config.ts";

export type Locals = Record<string, unknown>;
export type Metadata = Record<string, unknown>;
export type ViewConstr = new (mtime: number, path: string, uri: string, dirMtime: number | undefined, dirPath: string, dirUri: string, cached: boolean) => View;

const stack: Locals[] = [];

export function pushLocals(locals: Locals) {
    stack.push(locals);
}
export function popLocals() {
    stack.pop();
}
export function getLocals() {
    const res = stack.at(-1);
    if (res == null) throw new Error("getLocals() not called in a render context");
    return res;
}


export default abstract class View {
    public static cache = new Map<string, View>();

    private _cache?: string;
    private _metadata?: Metadata;
    private _src?: string;
    private _timer?: number;

    public get key() {
        return this.path;
    }
    public get isIndex() {
        return this.dirMtime != null;
    }
    public get hasCache() {
        return this._cache != null;
    }

    public get metadata() {
        if (this._metadata != null) return this._metadata;

        let res = this.onMetadata();
        if (typeof res === "string") res = parseYaml(res) as Metadata ?? {};
        if (typeof res !== "object") res = { data: res };
        if (this.cached) this._metadata = res as Metadata;

        return res;
    }

    public get src() {
        return this._src ??= Deno.readTextFileSync(this.uri);
    }
    public get title() {
        return String(this.metadata.title ?? this.path.split("/").at(-1));
    }
    public get date() {
        const rawDate = this.metadata.date as any;
        if (rawDate == null) return undefined;

        const res = new Date(rawDate);
        if (res == null) return undefined;
        else return res.toISOString().split("T")[0];
    }

    public setTimeout() {
        if (this._timer != null) clearTimeout(this._timer);
        if (config.maxCache != null) this._timer = setTimeout(() => {
            View.cache.delete(this.key);
            if (config.debug) console.log("FREE " + this.path);
        }, config.maxCache * 1000);

        return this;
    }
    public clearTimeout() {
        if (this._timer != null) clearTimeout(this._timer);
        return this;
    }

    protected abstract onMetadata(): Metadata | string;
    protected abstract onRender(locals: Locals, full: boolean): string;

    protected locals(locals: Locals) {
        return {
            title: config.defaultTitle,
            page: this.isIndex ? config.defaultIndex : config.defaultPage,
            ...locals,
            ...this.metadata
        };
    }

    public render(locals: Locals, full: boolean) {
        if (this._cache != null) return this._cache;

        pushLocals(locals);
        try {
            const res = this.onRender(locals, full);
            if (this.cached) this._cache = res;

            return res;
        }
        finally {
            popLocals();
        }
    }

    public constructor(
        public readonly mtime: number,
        public readonly uri: string,
        public readonly path: string,
        public readonly dirMtime: number | undefined,
        public readonly dirUri: string,
        public readonly dirPath: string,
        public readonly cached = true
    ) { }

}
