import config, { isIgnored } from "../config.ts";
import type { ViewConstr } from "./View.ts";
import { dirname, normalize, resolve } from "@std/path/posix";
import MarkdownView from "./MarkdownView.ts";
import PugView from "./PugView.ts";
import JSView from "./JSView.ts";
import { getPlugin } from "../plugins.ts";
import View from "./View.ts";

let getType = (ext: string) => {
    const types = {
        ".md": MarkdownView,
        ".pug": PugView,
        ".js": JSView,
        ".ts": JSView,
    };

    getType = ext => types[ext as keyof typeof types];
    return types[ext as keyof typeof types];
}

function tryStatFile(path: string) {
    try {
        return Deno.statSync(path);
    }
    catch {
        return undefined;
    }
}

function tryMtime(path: string | Deno.FileInfo | undefined) {
    if (typeof path === "string") {
        const stat = tryStatFile(path);
        if (stat?.mtime != null) return Number(stat.mtime);
        else return undefined;
    }
    else {
        if (path?.mtime != null) return Number(path.mtime);
        else return undefined;
    }
}

export function of(
    constr: ViewConstr,
    uri: string, path: string, stat: Deno.FileInfo | undefined,
    dirUri: string, dirPath: string, dirStat: Deno.FileInfo | undefined | boolean,
    cached = true, force = false
) {
    stat ??= tryStatFile(uri);
    const mtime = tryMtime(stat);

    if (stat?.isFile) {
        if (!force && isIgnored(path)) return undefined;
        if (dirStat === true) {
            dirStat = tryStatFile(dirUri) ?? undefined;
            if (dirStat == null) return undefined;
        }
        else if (dirStat === false) dirStat = undefined;
        if (typeof dirStat === "boolean") throw "fuck typescript";

        const index = dirStat != null;

        if (cached && View.cache.has(path)) {
            const res = View.cache.get(path)!;
            if (res.path === path && res.isIndex === index && res.mtime >= mtime!) {
                if (dirStat != null) {
                    if (res.dirMtime! >= Number(dirStat.mtime)) return res.setTimeout();
                }
                else return res.setTimeout();
            }
        }

        const res = new constr(mtime!, uri, path, tryMtime(dirStat), normalize(dirUri + "/"), normalize(dirPath + "/"), cached);

        if (cached) {
            const old = View.cache.get(path);
            if (old != null) old.clearTimeout();

            View.cache.set(path, res);
            return res.setTimeout();
        }

        return res;
    }
}

export function invoke(view: View, name: string, ...args: unknown[]) {
    if (config.libs == null) throw new Error("No 'lib' folder specified");
    return getPlugin(name).apply(view, args);
}

export function resolveView(path: string, cached = true, force = false): View | undefined {
    path = resolve("/" + path);

    const uri = resolve(config.views + path);
    let res;

    const dirStat = tryStatFile(uri);

    if (dirStat?.isDirectory) {
        res = [
            of(MarkdownView, uri + "/index.md", path, undefined, uri + "/", path + "/", dirStat, cached, force),
            of(PugView, uri + "/index.pug", path, undefined, uri + "/", path + "/", dirStat, cached, force),
            of(JSView, uri + "/index.js", path, undefined, uri + "/", path + "/", dirStat, cached, force),
            of(JSView, uri + "/index.ts", path, undefined, uri + "/", path + "/", dirStat, cached, force),
            of(MarkdownView, uri + ".md", path, undefined, uri + "/", path + "/", dirStat, cached, force),
            of(PugView, uri + ".pug", path, undefined, uri + "/", path + "/", dirStat, cached, force),
            of(JSView, uri + ".js", path, undefined, uri + "/", path + "/", dirStat, cached, force),
            of(JSView, uri + ".ts", path, undefined, uri + "/", path + "/", dirStat, cached, force),
        ].find(v => v != null);
        if (res != null) return res;

        if (config.defaultIndex != null) {
            const templUri = resolve(config.views!, config.defaultIndex);
            return [
                of(MarkdownView, templUri + ".md", path, undefined, uri + "/", path + "/", dirStat, cached, force),
                of(PugView, templUri + ".pug", path, undefined, uri + "/", path + "/", dirStat, cached, force),
                of(JSView, templUri + ".js", path, undefined, uri + "/", path + "/", dirStat, cached, force),
                of(JSView, templUri + ".ts", path, undefined, uri + "/", path + "/", dirStat, cached, force),
            ].find(v => v != null);
        }
    }
    else {
        return [
            of(MarkdownView, uri + ".md", path, undefined, dirname(uri) + "/", dirname(path) + "/", undefined, cached, force),
            of(PugView, uri + ".pug", path, undefined, dirname(uri) + "/", dirname(path) + "/", undefined, cached, force),
            of(JSView, uri + ".js", path, undefined, dirname(uri) + "/", dirname(path) + "/", undefined, cached, force),
            of(JSView, uri + ".ts", path, undefined, dirname(uri) + "/", dirname(path) + "/", undefined, cached, force),
        ].find(v => v != null);
    }

    return undefined;
}
export function resolveViewFromUri(uri: string, cached = true, index = true) {
    uri = resolve(uri);

    const views = resolve(config.views!);
    if (!uri.startsWith(views + "/")) return undefined;

    const getFile = (ext: string) => {
        const type = getType(ext);
        if (type == null) return undefined;

        const path = uri.slice(views.length, -ext.length);
        const dirPath = dirname(path);
        const dirUri = resolve(views + "/" + dirPath);

        const dirStat = tryStatFile(dirUri);

        if (dirStat?.isDirectory) {
            return of(type, uri, path, stat, resolve(views + "/" + path), path, dirStat, cached);
        }
        else {
            return of(type, uri, path, stat, dirUri, dirPath, undefined, cached);
        }
    };
    const getDir = (ext: string) => {
        const type = getType(ext);
        if (type == null) return undefined;

        let newUri = uri + "/index" + ext;

        let fileStat = tryStatFile(newUri);
        if (fileStat == null) {
            if (config.defaultIndex != null) {
                newUri = resolve(config.views + "/" + config.defaultIndex + ext);
                fileStat = tryStatFile(newUri);
            }

            if (fileStat == null) return undefined;
        }

        const path = uri.slice(views.length);
        const dirUri = uri;

        return of(type, newUri, path, fileStat, dirUri, path, stat ?? undefined, cached);
    };
    const getIndex = (ext: string) => {
        if (!index) return undefined;
        if (!uri.endsWith("/index" + ext)) return undefined;

        const type = getType(ext);
        if (type == null) return undefined;

        const dirUri = dirname(uri);
        const dirPath = dirUri.slice(views.length);
        const dirStat = Deno.statSync(dirUri);

        return of(type, uri, dirPath, stat, dirUri, dirPath, dirStat, cached);
    };

    const stat = tryStatFile(uri);
    let match;

    if (stat == null) return undefined;
    else if (match = uri.match(/index(\.(?:md|pug|js))$/)) return getIndex(match![1]);
    else if ((match = uri.match(/\.(?:md|pug|js)$/)) && stat.isFile) return getFile(match![0]);
    else if (stat.isDirectory) return getDir(".md") ?? getDir(".pug") ?? getDir(".js") ?? getDir(".ts");
    else return undefined;
}
export function renderView(view: string, locals = {}, full = false, force = false, passthroughLocals = false) {
    const res = resolveView(decodeURIComponent(view), true, force);
    if (res == null) return undefined;
    const cached = res.hasCache;
    return {
        text: res?.render({
            ...locals,
            ...(passthroughLocals ? {} : {
                dir: res.dirPath,
                dirUri: res.dirUri,
                path: res.path,
                uri: res.uri,
                isIndex: res.isIndex,
            })
        }, full),
        cached,
    };
}